if (typeof require !== 'undefined') {
 // for node.js
 var DictWrapper = require("./dict_wrapper.js").DictWrapper;
 var ddlm_dic = require("../ddl/ddlm_dic.js").ddlm_dic;
 var dREL_handler = require("../drel/drel_handler.js").dREL_handler;
 var drel = require("../drel/drel.js").parser; // the default name from jison

}

/*
 * Define a Meta Dictionary - a wrapper around potentially multiple
 * DDLm sub dictionaries, defined to implement the BaseDictWrapper interface.
 * 
 */
function MetaDict(environment) {
  this._logger = environment;
  this._enviro = environment;
  this._dict = { };  // these are loaded
  this._required = { };  // these dicts are needed but not loaded
  // check this._enviro.loadingDictQ for pending ...
  this._ddl_version ;
  this._import_hier = { };
  this._dict_parents = { };
  this._ddl_map = { };
  this._sub_cats = { };
  this.ddl = null;
}

MetaDict.prototype.log = function(data) {

  if (this._logger) this._logger.info(data);
}
MetaDict.prototype.debug = function(data) {
  if (this._logger) this._logger.debug(data);
}
MetaDict.prototype.info = function(data) {
  if (this._logger) this._logger.info(data);
}
MetaDict.prototype.warn = function(data) {
  if (this._logger) this._logger.warn(data);
}
MetaDict.prototype.error = function(data) {
  if (this._logger) this._logger.error(data);
}
MetaDict.prototype.fatal = function(data) {
  if (this._logger) this._logger.fatal(data);
}

MetaDict.prototype.reset = function() {
  for (var file in this._dict) {
     delete this._dict[file];
  }
  this._ddl_version = null;
}

MetaDict.prototype.wrapJSONDict = function(filename, json) {
  var wrapped =  new DictWrapper(json, this); // build hierarchy
  if (! this._ddl_version) {
    this._ddl_version = wrapped._ddl_version; //
    for (var key in wrapped._ddl_map) {
      this[key] = wrapped._ddl_map[key];
      this._ddl_map[key] = wrapped._ddl_map[key];
    }
    this._sub_cats = wrapped._sub_cats; // this is a very bad hack!
    if (wrapped._ddl_version == 'DDLm') {
      this.ddl = new DictWrapper( new ddlm_dic()._dict, this);    
//      alert("Loaded DDLm (as JSON)" );
    }
  }
  else if ( this._ddl_version != wrapped._ddl_version) {
     alert("Mismatched DDL versions of " + this._ddl_version + " and " + 
         wrapped._ddl_version + " for dict " + filename);
  }
  this._dict[filename] = wrapped; 
  if (filename in this._required) {
    delete this._required[filename];
    // if it was required to be loaded, it no longer is.
  }
  return this;
}

MetaDict.prototype.getDict = function(filename) {
  return this._dict[filename];
}



MetaDict.prototype.getHTML = function(docroot) {
  var HTML = docroot.createDocumentFragment();
  for (var filename in this._dict) {
    var html = this.getHTML4dict(docroot, filename); 
    HTML.appendChild(html);
  }
  return HTML;
}

MetaDict.prototype.getHTML4dict = function(docroot, filename) {
  //var convertor = new JS_DIC2HTML(docroot, filename,'dic:');
  var convertor = new HierarchyView(docroot, filename,'dic:');
  convertor._logger = this._logger;
  var html = convertor.renderModel(this.ddl, this._dict[filename]); 
  return html;
}


MetaDict.prototype._getImportFileList = function(imprt) {
  //  imprt is a list of ref hashes
  var files = {};
  for (var i = 0; i < imprt.length; i++) {
    var ref = imprt[i];
    var file_name = ref['file'];
    files[file_name] = 1;
  }
  return files; // return unique list as a hash
}


MetaDict.prototype._getFileImportNames4Dict = function(dict_file_name) {

  this.info("check  " + dict_file_name + " for import tag:  " + this._defn_import  );
    
  var dict = this._dict[dict_file_name];
//  var import_items = { };
//  var import_cats = { };
  var files = {}; 
  for (var cat_name in dict._categories ) {
    var cat = dict._categories[cat_name];
    var imprt = cat.getAttribute(this._defn_import);
    if (imprt) {
      var file_hash = this._getImportFileList(imprt);
      for (var f in file_hash) {
        files[f] = file_hash[f];
      }
    }
  }
  for (var item_name in dict._items ) {
    var item = dict._items[item_name];
    var imprt = item.getAttribute(this._defn_import);
    if (imprt) {
      var file_hash = this._getImportFileList(imprt);
      for (var f in file_hash) {
        files[f] = file_hash[f];
      }
    }
  }
  return files;

}

MetaDict.prototype.buildFileImportList = function(dict_file_name) {
  if (this._ddl_version != 'DDLm') return;

  var import_files = this._getFileImportNames4Dict(dict_file_name);
  this._import_hier[dict_file_name] = import_files; // record to build dict hier

  var extra = [];
  for (var filename in import_files) {
    if (filename in this._dict ) continue;
    if (this._enviro.loadingDictQ(filename)) continue; // in pending
    this._required[filename] = import_files[filename];  // record needed    
    extra.push(filename);
  }
  if (extra.length) {
    this.warn(dict_file_name + " needs added imports of" + extra );
  }
}


MetaDict.prototype.resolveRequiredImports = function() {
  // remember, this is invoked thru an asynchronous file load (ala AJAX)
  var origList = {};
  var dlist = [];
  for (var filename in this._required) {
    origList[filename] = 1;
    dlist.push(filename);
  }
  for (var filename in origList) {  // if required and not in pending
    if ((filename in this._required)  &&
         ! this._enviro.loadingDictQ(filename)) {
      this._enviro.userLoadsDict(filename);  // then request a load.
      break; // only load one. 
      // that in turn will come back to load another, if needed
    }
  }

  // this will be true after the final dictionary file dependency was loaded
  if (dlist.length == 0) {
    if (this._ddl_version == 'DDLm') { 
      this._ultimateImportResolution();
      this._buildDictHierarchy();  // reverse lookup for dict imports
    }
    // alert user we are ready to go
    alert("Dictionaries all loaded and resolved!");
  }
}


MetaDict.prototype._ultimateImportResolution = function() {
  try {
  for (var src_dict_file_name in this._dict) {
            this.info("Post loading resolving imports for " +
                 src_dict_file_name );
    var dict = this._dict[src_dict_file_name];
    var saves = [dict._categories, dict._items];//anything useful
    for (var j=0; j<2; j++) { 
      var grp = saves[j];
      for (var saved in grp ) {
        var src_save = grp[saved];
        if (src_save instanceof Array) continue; // schitzophrenic aliases :-/
        var imprt = src_save.getAttribute(this._defn_import);
        if (! imprt) continue;
        var imports = [ ];
        for (var i = 0; i < imprt.length; i++) {
          var ref = imprt[i];
          var trg_dict_file_name = ref['file'];
          var save_frame_name = ref['save'];
          var save;
          var trg_dict = this._dict[trg_dict_file_name];
          if (trg_dict) {
            save = trg_dict.getCategory(save_frame_name.toLowerCase());
            if (! save) {
              save = trg_dict.getItem(save_frame_name.toLowerCase());
            }
            if (! save) {
              save = trg_dict._generics[save_frame_name.toLowerCase()];
            }
          }
          if (save) {
            imports.push(save) ; // make cross dict link
          }  
          else {
            this.warn("Missing save frame  " + save_frame_name + 
                " while importing from "  + trg_dict_file_name + " for " +
                src_save._name + " in " + src_dict_file_name );
          }
        }
        if (imports.length >1) {
          src_save._generic = imports;
        } else if (imports.length ==1) {
          src_save._generic = imports[0];
        }
      } // end of group
    } // end of groups
  } // end of dict
  } catch (e) {
    alert(e);
  } 
}

MetaDict.prototype._buildDictHierarchy = function() {
  // make a hash of who includes 'that' dictionary
  var copy = {};
  for (var filename in this._import_hier) {
    copy[filename] = []; 
  }
  for (var filename in this._import_hier ) {
    var hash = this._import_hier[filename];
    for (var impfile in hash) {
      copy[impfile].push(filename); 
    }
  }
  this._dict_parents = copy;
}


MetaDict.prototype.verifyAllImports = function() {
  if (this._ddl_version != 'DDLm') return;
  for (var filename in this._dict) {
    this.buildFileImportList(filename) ;
  }
  
  alert("need to load these too : " + extra);
  for (var i = 0; i < extra.length; i++) {
    this._enviro.userLoadsDict(extra[i]);
  }

}




/*
 * Common interface methods for all derived wrappers
 *
 */
MetaDict.prototype.getItem = function(name) {
  // loop for all dictionaries
  for ( var dict_file in this._dict) { 
    var item = this._dict[dict_file]._items[name];
    if (item) return item;
  }
}
MetaDict.prototype.getCategory = function(name) {
  // loop for all dictionaries
  for ( var dict_file in this._dict) { 
    var item = this._dict[dict_file]._categories[name];
    if (item) return item;
  }
}
MetaDict.prototype.hasCategory = function(name) {
  return this.getCategory(name);
}

MetaDict.prototype.Categories = function() {
  // loop for all dictionaries
  var cats = {};
  for ( var dict_file in this._dict) { 
    var dict = this._dict[dict_file];
    for (var cat in dict._categories) {
      cats[cat] = 1;
    }
  }
  return cats;
}
MetaDict.prototype.getCategories = function() {
  return this.Categories();
}


MetaDict.prototype.hasItem = function(name) {
  return this.getItem(name);
}

MetaDict.prototype.getFunction = function(name) {
  // loop for all dictionaries
  for ( var dict_file in this._dict) { 
    var func = this._dict[dict_file]._functions[name];
    if (func) return func;
  }
}

MetaDict.prototype.hasFunction = function(name) {
  return this.getFunction(name);
}

MetaDict.prototype.getDDLVersion = function() {
  for ( var dict_file in this._dict) { 
    var item = this._dict[dict_file]._categories[name];
    if (item) return item;
  }
}

MetaDict.prototype.getItemPath = function(item_name, prefix) {
  // loop for all dictionaries
  for ( var dict_file in this._dict) { 
    var dict = this._dict[dict_file];
    var item = dict._items[item_name];
    if (item) {
      var plist = dict.getItemPath(item_name,prefix);  
      var includee = this._dict_parents[dict_file];
      while (includee && includee.length > 0) {
        var catname = plist[plist.length -1]; 
        var pdict = includee[0]; 
        var include_dict = this._dict[pdict];
        var parnt = include_dict._categories[catname]; 
        while (parnt._category) {
          plist.push(parnt._category._name.toLowerCase());
          parnt = parnt._category;
        }
        includee = this._dict_parents[pdict];
      }
//    this.info(plist);
      plist.push(dict.getName());
      return  plist;
    }
  }
}

MetaDict.prototype.getCategoryHierarchy = function(cat_name) {
  // loop for all dictionaries
  for ( var dict_file in this._dict) { 
    var dict = this._dict[dict_file];
    var cat = dict._categories[cat_name];
    if (cat) {
      var plist = dict.getCategoryPath(cat_name);  
      plist.shift(); // rip cat_name off bottom of stack
      var includee = this._dict_parents[dict_file];
//      alert(includee);
      while (includee && includee.length > 0) {
        var catname = plist[plist.length -1]; 
        var pdict = includee[0]; 
        var include_dict = this._dict[pdict];
        var parnt = include_dict._categories[catname]; 
        while (parnt._category) {
          plist.push(parnt._category._name.toLowerCase());
          parnt = parnt._category;
        }
        includee = this._dict_parents[pdict];
      }
//    this.info(plist);
//      plist.push(dict.getName());
      return  plist;
    }
  }
}


MetaDict.prototype.getItemDictPrefix = function(item_name) {
  // loop for all dictionaries
  for ( var dict_file in this._dict) { 
    var dict = this._dict[dict_file];
    var item = dict._items[item_name];
    if (item) {
      return 'dic:' + dict_file;  // hack
      //return 'dic:' + dict_file + dict.getName() ;  // hack
    }
  }
}


MetaDict.prototype.getJoinedCats = function(cat_name) {
  // loop for all sub dictionaries
  var set = {}
  for ( var dict_file in this._dict) { 
    var dict = this._dict[dict_file];
    var addset = dict.getJoinedCats(cat_name);
    if (addset) {
      for ( var key in addset) { 
        set[key] = addset[key];
      }
    }
  }
  return set;
}

MetaDict.prototype.getCatDefaults = function(cat_name) {
  // loop for all dictionaries
  for ( var dict_file in this._dict) { 
    var dict = this._dict[dict_file];
    var deflt = dict.getCatDefaults(cat_name);
    if (deflt) {
      return deflt;
    }
  }
}


MetaDict.prototype.getMethodString = function(defn, klass) {
  var methodClass = defn.getAttribute(this._method_purpose);
  var methodString;
  if ( (! methodClass && klass == 'evaluation') ||
       ( methodClass && methodClass.toLowerCase() == klass)) {
    methodString = defn.getAttribute(this._method_expression);
  }
  if (!methodString) {
    var methodData = defn.getAttributeLoop(this._method_expression);
    if (methodData) {
      var tags = methodData.map;
      var rows = methodData.rows;
      for (var i = 0; i< rows.length; i++) {
        if (rows[i][tags[this._method_purpose]].toLowerCase() == klass) {
          methodString = rows[i][tags[this._method_expression]];
          break;
        }
      }
    }
  }
  if (methodString) {
     methodString += "\n";
  }
  return methodString;
}

MetaDict.prototype.getDefnFuncDepends = function(defn, purpose, lib) {
  if (defn.hasCachedMethod(purpose)) {
    return defn.getCachedMethodDepends(purpose);
  }  
  var func = this.getDefnFunction(defn, purpose, lib);
  if (func) {
    return defn.getCachedMethodDepends(purpose);
  } 
  return null;
}

MetaDict.prototype.getDefnFunction = function(defn, purpose, lib) {
  if (defn.hasCachedMethod(purpose)) {
    return defn.getCachedMethod(purpose);
  }  
    
  var string = this.getMethodString(defn, purpose);
  if (! string ) {
    defn.cacheMethod(purpose, null, null);
    return null;
  }

  var func = null;
  var deps = null;
  try {
    var funcdep = this.compileMethodString(defn._name, string, purpose, lib);
    func = funcdep[0];
    deps = funcdep[1];
  } catch (e) {
    alert("failed to compile " + purpose + " method " + defn._name + "\n"
             + string);
    this.error("failed to compile defn meth. " + defn._name + "\n"
           + string);
    this.error(e);
  }
  
  defn.cacheMethod(purpose, func, deps);
  return func;
}


MetaDict.prototype.compileMethodString = function(defnName, methodString, klass, lib) {
  // generate Abstract Syntax Tree from dREL source
  var drel_parser = drel; // from drel.js
 this.warn("parsing method " + defnName + " " + klass);
  var ast = drel_parser.parse(methodString);
  // Compile the ast to javascript
  var dREL = new dREL_handler();
  //var jsMethodSource = dREL.compile(name, ast, this._dict, this._lib, methklass);
 this.warn("compiling  source");
  var cmp = dREL.compileDependencies(defnName, ast, this, lib, klass);
  var jsMethodSource = cmp[0];
  var depends = cmp[1];
  this.warn("Transmogrified: " +defnName);
  this._dumpDep(depends);
  this.info(jsMethodSource);
  var func = new Function('Ctxt','Library', jsMethodSource);
  return [ func, depends];
}

MetaDict.prototype.compileMethodDepends = function(defnName, methodString, klass, lib) {
  // generate Abstract Syntax Tree from dREL source
  var drel_parser = drel; // from drel.js
  var ast = drel_parser.parse(methodString);
  // Compile the ast to javascript
  var dREL = new dREL_handler();
  //var jsMethodSource = dREL.compile(name, ast, this._dict, this._lib, methklass);
  var cmp = dREL.compileDependencies(defnName, ast, this, lib, klass);
  var jsMethodSource = cmp[0];
  var depends = cmp[1];
  this.warn("Transmogrified: " +defnName);
  this._dumpDep(depends);
  this.info(jsMethodSource);
//  var func = new Function('Ctxt','Library', jsMethodSource);
//  return func;
  return depends;

}

MetaDict.prototype._dumpDep = function(deps) {
  var str = "Dependencies\n"
  for (var key in deps) {
    str = str + key + "\n";
    var type = deps[key];
    for (var dep in type) {
      str = str + "   " +dep + " " + type[dep] + "\n";
    }
  }
  this.info(str);
}


// for node.js
if (typeof require !== 'undefined' && typeof exports !== 'undefined') {
  exports.MetaDict = MetaDict;
}
