                          PROGRAM TAGXREF
c
c      Version 19 April 2013   Author: sydney.hall@uwa.edu.au
c
c >>  Program to X-reference the occurrence of data TAGs in text files  <<
c >>  corresponding to those defined in a DDLm dictionary file.         <<
c >>  An important by-product of this process is the expansion of the   <<
c >>  dictionary and the text input files (xridic & xritex) to include  <<
c >>  the definition information requested by _import.get statements.   <<
c >>  The expanded files are xrodic and xrotex respectively.            <<
c
c ..........................................................................
c
c   The input text file to be checked is opened as 'xritex'.
c   The input x-check dictionary file is opened as 'xridic'.
c
c   The 'xridic' file is read by LOADDICT and copied to the file 'xrodic'
c   During this copy all 'import.get' statements are executed and the
c   imported lines are added.
c
c   The 'xritex' file is read by XCHECK and copied to the file 'xrotex'
c   During this copy, if the text file is recognised as a dictionary, all 
c   'import.get' statements are executed and the imported lines are added.
c   At the same time 'xrotex' is scanned for tags that are defined in
c   'xrodic' and the line numbers (wrt the file 'xrotex') are recorded
c   in the file 'xrlis'.
c
c   Note that the variable 'maxc' is the maximum number of characters in
c   a single input and output line - the preset value is 125.
c
c  >>>>>> PROGRAMMERS see additional notes at the end of the source.
c
c .............................installation notes.............................
c
c  Input dictionary file    'xridic'    device 11+ (routine LOADDICT)
c  Output dictionary file   'xrodic'    device 11  (routine LOADDICT)
c  Input text check file    'xritex'    device 11+ (routine XCHECK)
c  Output text check file   'xrotex'    device 11  (routine XCHECK)
c  Output list of tag lines 'xrlis'     device  4  (routine REPORT)

c  ERROR/DUMP printouts    <stdout>     device  6  (all routines)
c  IGNORED tag list       'XTAGS'       device  3  (routine LOADDICT)
c  DUMP codes list        'XREF'        device  7  (routine LOADDICT)
c
c  'XREF' file enables the user to control both the mode of searching
c  'xrotex' for tags, and diagnostic printouts in the event of a illegal
c   dictionary files structure (e.g. unsatisfied import statement). The
c   control codes concatenated on the first line of 'XREF' are:
c   ;   = also scan for tags inside the ;-delimited text of 'xrotex'
c   h   = DUMP history of the import occurrences and nesting
c   d   = DUMP details of the import processing for xridic => xrodic
c   r   = DUMP details of all lines read from  xridic and xritex
c   w   = DUMP details of all lines written to xrodic and xrotex
c   s   = DUMP details of all token/string parsing (verbose!)
c
c ..........................................................................
c
      include 'tagxref.cmn'
      character pcodes*20,c*1
c
c            v----------------------max chars input text file
      maxc=125 
      maxnam=5000
      case=ichar('a')-ichar('A')
      tab=char(05)
      if(case.ge.0) tab=char(09)
c
c     control scan & diagnostic prints
      b=.false.
      h=.false.
      d=.false.
      w=.false.
      r=.false.
      s=.false.
      open(7,file='XREF',status='old',form='formatted')
      read(7,'(a20)',end=50) pcodes
      do 10 i=1,20
      c=pcodes(i:i)
      if(c.eq.';') b=.true.
      if(c.eq.'h') h=.true.
      if(c.eq.'d') d=.true.
      if(c.eq.'w') w=.true.
      if(c.eq.'r') r=.true.
      if(c.eq.'s') s=.true.
10    continue
      close(7)
c
      call date_and_time(DATE=date,TIME=time)
      date=date(1:4)//'-'//date(5:6)//'-'//date(7:8)
      time=time(1:2)//':'//time(3:4)
c
50    write(*,'(/10(1H<),3a,10(1H>)/)')
     * ' TAGxref - X-ref text file with defined tags at ',
     * time(1:6),date(1:11)
c
      open(4,file='xrlis',status='new',form='formatted')
c
c                        Load the dictionary and write xrodic
      call LOADDICT
c
c                        Load the text file to be scanned
      call XCHECK
c
c                        Report tags found in xrotex wrt xrodic
      call REPORT
c
      close(7)
      close(8)
      close(11)
100   continue
      end
c
c***************************************************************************
c     LOADDICT transfer xridic with imported defs to temp file. Scan temp file 
c     for defs and place these in dnam(),lnam().                        
c***************************************************************************
c
      subroutine LOADDICT
      include 'tagxref.cmn'
      integer i,j
      character nona*21
      data nona/',([{</|".?!;:)]}>/-=|'/
c
c-------------------------------------- reset the dictionary tag list
      wcan=0
      text=0
      nnam=0
      nfil=0
      mfil=0
      stri=' '
      mdef=0
      nbuf=0
      ntok=0
      icach(1)=1
      do 3 i=1,100
3     line(i)=0
      do 5 i=1,25
      slev(i)=0
      nimp(i)=0
      mimp(i)=0
      devn(i)=i+10
      do 5 j=1,25
      do 5 k=1,5
      dicl(i,j,k)=0
5     dicn(i,j,k)=' '
      do 10 i=1,maxnam
      lnam(i)=0
10    dnam(i)=' '
c-------------------- open xridic (nfil=1); op to expanded xrodic (file7) 
      fram=' '
      dupl=' '
      miss=' '
      mode=' '
      filn='xridic'
      call newfil
      writ=1
      open(7,file='xrodic',status='new',form='formatted')
c----------------------------- expand xridic with imports to xrodic file  
20    call GETTOK
      if(ieof.ne.0)                             goto 50 
      ntok=ntok+1
      if(ntok.gt.10)                            goto 30
      if(type.ne.'name')                        goto 20
      if(stri(1:nchr).ne.'_dictionary.title')   goto 22
      call GETTOK
      dtitl=stri(1:nchr)
22    if(stri(1:nchr).ne.'_dictionary.version') goto 24
      call GETTOK
      dvers=stri(1:nchr)
24    if(stri(1:nchr).ne.'_dictionary.date')    goto 20
      call GETTOK
      dddat=stri(1:nchr)
      write(*,'(16x,4a/)') '>>>> xref dictionary is: ',dtitl,dvers,dddat
      goto 20
C
30    if(type.ne.'impt') goto 20
      call IMPORT
      if(ieof.ne.0)      goto 50 
      goto 20
c
c-------------------------------- read xrodic, store defs in dnam()
50    writ=0
      ieof=0
      close(7)
      open(7,file='xrodic',status='old',form='formatted')
      ndev=7
      line(7)=0
      ichr=maxc
c
      call PUTDEF
c
Cdbg  write(6,'(a,i5,2x,a)') (' cache ',icach(i),ncach(i),i=1,mdef+1)
Cdbg  write(6,'(i5,2x,a)') (i,cacbuf(i),i=1,nbuf)                   
Cdbg  write(6,'(//a)') '>>>>> dictionary tags stored are:'
Cdbg  write(6,'(2i4,4x,2a40)') (i,lnam(i),inam(i),dnam(i),i=1,nnam)
Cdbg  write(6,'(///)')
      return
      end
c
c
c***************************************************************************
c     XCHECK reads input text file 'xritex' and generates the file 'xrotex'
c     which contains any imported data if the xritex text file is also a
c     dictionary. Then reads xrotex searching for data names and compares
c     these with those defined in the xrodic file (the expanded xridic file)
c***************************************************************************
c
      subroutine XCHECK
      include 'tagxref.cmn'
      integer i,j,n,dicy
      character f*1,f1*1,f2*1
      character tail*17,c*1
      data tail/' ,.?!;:)]}>/-=|" '/
      character*80 xtag(1000),name,mame
      integer      xlen(1000)
c
      tail(1:1)=''''
      tail(17:17)=tab
      text=1
      nfil=0
      mfil=0
      nhis=0
      mnam=nnam+1
      do 5 i=1,25
      slev(i)=0
      nimp(i)=0
      mimp(i)=0
      devn(i)=i+10
      do 5 j=1,25
      do 5 k=1,5
      dicl(i,j,k)=0
5     dicn(i,j,k)=' '
      do 8 i=1,1000
      do 8 j=1,25
      do 8 k=1,5
8     hist(i,j,k)=' '
c
c  -------------------- open xritex (nfil=1); op expanded xrotex (file7)  
c                       assume not dict (dicy=0); reset to =1 if it is
      ntok=0
      dicy=0
      fram=' '
      dupl=' '
      miss=' '
      mode=' '
      filn='xritex'
      call NEWFIL
      writ=1
      open(7,file='xrotex',status='new',form='formatted')
c
c  ------------------ read xritex and write xrotex (with imports if dic)
c
10    call GETTOK
      if(ieof.ne.0)                             goto 30
      ntok=ntok+1
      if(ntok.gt.10)                            goto 20
      if(type.ne.'name')                        goto 10
      if(stri(1:nchr).ne.'_dictionary.title')   goto 12
      call GETTOK
      ttitl=stri(1:nchr)
      dicy=1
12    if(stri(1:nchr).ne.'_dictionary.version') goto 14
      call GETTOK
      tvers=stri(1:nchr)
14    if(stri(1:nchr).ne.'_dictionary.date')    goto 10
      call GETTOK
      tddat=stri(1:nchr)
      write(*,'(16x,4a/)') '>>>> checking text dic: ',ttitl,tvers,tddat
      goto 10
C
20    if(dicy.eq.0)                             goto 10
      if(type.ne.'impt')                        goto 10
      call IMPORT
      if(ieof.ne.0)                             goto 30
      goto 10
c
30    ieof=0
      writ=0
      ndev=7
      close(7)
      open(7,file='xrotex',status='old',form='formatted')
      line(7)=0
      ichr=maxc
c
      if(.not.h) goto 34
      write(6,'(/a)') ' IMPORT HISTORY......'
      do 33 i=1,nhis
      write(6,'(/a,i3,a)') ' POP',i,
     *' NFIL     FILE           FRAME         IF_DUPL  IF_MISS  MODE'
      j=1
32    write(6,'(i11,6x,2a15,3a10)') j,(hist(i,j,k),k=1,5)
      j=j+1
      if(hist(i,j,2)(1:1).ne.' ') goto 32 
33    continue
c
c
c -------------------------------- load XTAGS (tags to be ignored)
34    xcnt=0
      open(3,file='XTAGS',status='old',form='formatted')
      do 40 i=1,1000
      read(3,'(a)',end=45) name
      if(name(1:1).eq.'#') goto 40
      xcnt=xcnt+1
      do 35 j=1,60
      if(name(j:j).eq.' ') goto 36
35    continue
36    xlen(xcnt)=j-1
      xtag(xcnt)=name(1:j-1)
40    continue
45    close(3)
c
c --------------------------------- enter xrotex to be checked
50    call GETTOK
      if(ieof.ne.0)        goto 1000
      if(type.eq.'name')   goto 100
      if(type.eq.'char')   goto 200
      if(type.eq.'text')   goto 200
      goto 50
c
c ---------------------------- this is an attribute tag!
100   continue
      mame=stri(1:nchr)
      call lcase(mame,name,m)
      i=1
cdbg  write(6,'(a,3i5,1(x,a))')'>>>>>>>>>>>>',m,dicy,semi,name(1:m)
      if(dicy.eq.0)  goto 260
      if(semi.eq.1)  goto 260
      if(stri(1:25).ne.'_dictionary_audit.version')  goto 260
      goto 1000
c
c ---------------------------- this is a string; check if tag?
200   continue
      mame=stri(1:nchr)
      if(stri(1:1).eq."'") goto 210
      if(stri(1:1).ne.'"') goto 220
210   mame=stri(2:nchr)
220   call lcase(mame,name,m)
cdbg  if(name(m:m).eq.']') write(6,'(2(x,a))') name(1:m),mame(1:m)
      if(name(1:1).ne.'_') goto 50
      if(name(2:2).eq.'_') goto 50
c
      do 230 j=m,1,-1
      if(index(tail,name(j:j)).ne.0) goto 240
230   continue
      goto 250
240   name=name(1:j)
      m=j
250   if(m.lt.2) goto 50
cdbg  write(6,'(a,3i5,2(x,a))')'C',i,j,m,name(1:m),stri(1:nchr)
c
c --------------------------- this is a tag; remove trailing junk
260   if(m.gt.60)
     * call err(0,' tag in text > 60 chars',stri(i:j))
300   if(name(m:m).eq.'_')     goto  50
      if(name(m:m).eq.'.')     goto 305
      if(name(m:m).eq.',')     goto 305
      if(name(m:m).eq.'*')     goto 305
      if(name(m:m).eq.')')     goto 305
      if(name(m:m).eq.'}')     goto 305
      if(name(m:m).eq.'"')     goto 305
      if(name(m:m).eq."'")     goto 305
      goto 310
305   m=m-1
      goto 300
310   if(name(m-1:m).eq.'[]')  goto 50
c --------------------------------------- search if tag in list     
320   nnn=1
      do 325 i=1,nnam
      if(name(1:m).eq.dnam(i)) goto 330
325   continue
      nnn=0
c ---------------------------------------- check if tag to be ignored
330   do 340 n=1,xcnt
      if(xlen(n).ne.m)         goto 340
      if(xtag(n).eq.name(1:m)) goto 350
340   continue
      if(nnn.eq.1)        goto 550
      goto 500
350   npos(i,1)=-1
      goto 50
c ---------------------------------------- new data name add to list
500   nnam=nnam+1
      if(nnam.le.maxnam) goto 510
      call err(1,'data name table exceeded. current max is','5000')
510   dnam(nnam)=name(1:m)
      inam(nnam)=mame(1:m)
      lnam(nnam)=m
cdbg  write(6,'(3i5,5x,2a)') m,nnam,line(ndev),stri(1:nchr),name
      do 520 j=1,100
520   npos(nnam,j)=0
      i=nnam
c --------------------------------------- add line numbers to npos()
550   if(npos(i,1).eq.99) goto 50
      if(npos(i,1).eq.-1) goto 50
      npos(i,1)=npos(i,1)+1
      n=npos(i,1)
      if(n.le.99)    goto 560
      write(6,'(a,a)') 'warning: line table truncated for ',mame
560   npos(i,n+1)=line(ndev)
      go to 50
c
c------------------------------------------- XTEXT is complete
1000  close(7)
      return
      end
c
c
c***************************************************************************
c     REPORT prints a summary of the line occurrences for each data name
c     in the dictionary and the text file itself (if it does not appear
c     in the dictionary). this summary is output as the file 'xrlis'.
c***************************************************************************
c
      subroutine REPORT
      include 'tagxref.cmn'
      integer i,j,n
      character*1 z
c
      n=0
      do 10 i=1,mnam-1
      if(lnam(i).lt.0) inam(i)(1:1)=' '
      if(npos(i,1).eq.0) goto 10
      n=n+1
10    continue
c
      if(b)write(*,'(//a)')' ******  ;-delimited text searched also'
      write(*,'(//35h        tags defined in xrodic file,i5)') mnam-1
      write(*,'(35h     xrotex tags NOT in xrodic file,i5)') nnam-mnam+1
      write(*,'(35h     xrotex tags in the xrodic file,i5)') n
      write(*,'(35h         ignored tags in XTAGS file,i5)') xcnt
      write(*,'(/5x,a)') ">> Note: see 'xrlis' for list of tags found"

      write(4,'(1x,10(1h>),3a,10(1h<)/)')
     *' TAGxref - List of X-referenced TAGs ',time(1:6),date(1:11)
      write(4,'( 7x,4a)') '>>>> xref dictionary is: ',dtitl,dvers,dddat
      write(4,'( 7x,4a/)') '>>>> checking text dic:  ',ttitl,tvers,tddat
      if(b)write(4,'(/a/)')' ******  ;-delimited text searched also'
      write(4,'(/35h        tags defined in xrodic file,i5)') mnam-1
      write(4,'(35h     xrotex tags NOT in xrodic file,i5)') nnam-mnam+1
      write(4,'(35h     xrotex tags in the xrodic file,i5)') n
      write(4,'(35h         ignored tags in XTAGS file,i5)') xcnt
      write(4,'(/5x,a)') ">> Note: see 'xrlog' for log x-ref process"
c
      write(4,'(//a,16(1h-),a/)') ' Text file tags NOT in dictionary ',
     * ' line numbers in xrotex file -----------'
      do 50 i=mnam,nnam
      n=npos(i,1)
      if(n.gt.0) goto 20
      if(n.lt.0) goto 15
      write(4,'(1x,a,2x,a)') inam(i),'- not present in xrotex'
      goto 50
15    write(4,'(1x,a,2x,a)') inam(i),'- found but ignored'
      goto 50
20    if(n.gt.5) goto 40
      write(4,'(1x,a51,5i6            )') inam(i),(npos(i,j+1),j=1,n)
      goto 50
40    write(4,'(1x,a51,5i6,/,(52x,5i6))') inam(i),(npos(i,j+1),j=1,n)
50    continue
c
      write(4,'(//a,20(1h-),a/)') ' Text file tags in dictionary ',
     * ' line numbers in xrotex file -----------'
      do 100 i=1,mnam-1
      z=' '
      if(lnam(i).lt.0) z='@'
      n=npos(i,1)
      if(n.gt.0) goto 60
      if(n.lt.0) goto 55
      write(4,'(1x,a,2x,a)') inam(i),'- not present in xrotex'
Cdbg  write(4,'(1x,a,2i6)')   dnam(i),npos(i,1),npos(i,2)
      goto 100
55    write(4,'(1x,a,2x,a)') inam(i),'- found but ignored'
      goto 100
60    if(n.gt.5) goto 70
      write(4,'(1x,a50,a1,5i6           )')inam(i),z,(npos(i,j+1),j=1,n)
      goto 100
70    write(4,'(1x,a50,a1,5i6,/(52x,5i6))')inam(i),z,(npos(i,j+1),j=1,n)
100   continue
c
cdbug write(6,'(i5,5x,a)') (icach(i),ncach(i),i=1,mdef+1)
      close(4)
      return
      end
c
c
c***************************************************************************
C
c     PUTDEF - Read the expanded xrodic file and store dict names from 
c              definition.id and alias.definition_id
c              Place them in "nami" (as input) and "name" (lower case)
c              Add to dictionary list as dnam() lower case name
c                                        inam() input name
c                                        lnam() length of name (- if alias)
c    
c***************************************************************************
c
      subroutine PUTDEF
c
      include 'tagxref.cmn'
      integer i
      character c*1, nona*21
      character*60 name,nami
      data nona/',([{</|".?!;:)]}>/-=|'/
c
      if(d)write(6,'(a)')'***start PUTDEF'
      iatt=0
c             -------------------------- read the xrodic file
10    call GETTOK
      if(ieof.ne.0)       goto 100
      if(iatt.ne.0)       goto 40
      if(type.eq.'impt')  goto 90         
      if(type.ne.'name')  goto 10
      if(stri(1:nchr).ne.'_definition.id')      goto 30      
      iatt=1
      goto 10
30    if(stri(1:nchr).ne.'_alias.definition_id')goto 35
      iatt=2 
      goto 10
35    iatt=0
      goto 10
c
c  --------------------- data name found; is it a definition name
40    if(stri(2:2).eq.'_') goto 45
      if(iatt.eq.2)        goto 45
      iatt=0
      goto 10
45    if(nchr.gt.62)
     * call err(0,'dictionary data name > 60 chars',stri(2:nchr-1))
      nami=stri(2:nchr-1)
      call lcase(nami,name,l)
c
c---------------------------------------test if tag or duplicate
      if(index(nona,name(2:2)).gt.0) goto 10
      do 50 i=1,nnam
      if(name.eq.dnam(i))call err(1,'duplicate definition  ',name) 
50    continue
c -------------------------------------- new tag add to list dnam()
      nnam=nnam+1
      if(nnam.gt.maxnam) call err(1,'dictionary tags exceed 5000','')
      dnam(nnam)=name
      inam(nnam)=nami
      lnam(nnam)=nchr-2
      if(iatt.eq.2) lnam(nnam)=-lnam(nnam)
      if(d)write(6,'(a,a30)')'}}}}}}}}}}}} PUTDEF   ',name
      goto 10
c
90    call err(1,'unexpected import in odict file','')
100   continue
cdbug write(6,'(2i3,2a40)') (i,lnam(i),inam(i),dnam(i),i=1,nnam)
      return
      end
c
c
c***************************************************************************
C
c     IMPORT imports the dictionary definition blocks as designated
c     by the "_import.get" attribute .....
c
c***************************************************************************
c                         file  'filename/URI of source dictionary'
c                         save  'save framecode of source definition'
c                         mode  '"full" with frame; "cont" default'
c                         dupl  'option for duplicate entries'
c                         miss  'option for missing duplicate entries'
c
c     _import.get = [{"file":i.file_id,"save":i.frame_id,"mode":i.mode,
c                     "dupl":i.if_dupl,"miss":i.if_miss}]
c
c---------------------------------------------------------------------------
c
      subroutine IMPORT
      include 'tagxref.cmn'
c
c                        get import args from list [{..},....]
      call GETVAL
c                        reset the import list pointer
100   call NEXIMP
      if(next.eq.1)       goto 200
      if(slev(nfil).gt.0) goto 400
      goto 800
c                        if there is a cache, search & retrieve
200   call GETCAC
      if(cacy.eq.1) goto 100
c                        not in cache find frame in import file
300   call NEXFRM
c                        get the import frame and copy to xrodic/xrotex
400   call GETFM1
      if(ieof.ne.0) goto 1000
c                       tabulate cache pointers for any new lines
      call TABCAC
      goto 100
c                         close import file nfil and decrement
800   if(nfil.le.1) goto 1000
      call CLOSEF
      goto 100
c
1000  writ=1
      if(d)write(6,'(2a,a30)')'\\\\\\  exit IMPORT ',mode,fram
c
      return
      end
c
c
c***************************************************************************
c
c    CLOSEF - close the current import file
c
c***************************************************************************
c
      subroutine CLOSEF
      include 'tagxref.cmn'
c
      if(mode.eq.'full') call WRITLN(ibuf,mchr)
      ibuf=' '
      cach=0
      writ=0
      close(ndev)
      if(h)write(6,'(8(2h--),a,i3,x,a)')'close',nfil
      nfil=nfil-1
      ndev=devn(nfil)
      if(d)write(6,'(8(2h- ),a,2i3)') ' next file',nfil,slev(nfil)
c
      continue
      return
      end
c
c***************************************************************************
c
c     GETVAL - get the arguments from the import.get list [{..},...]   
c              store these in dicn() strings file,fram,dupl,miss,mode
c                             dicl() lengths of these strings
c              nimp(),mimp() are the max, count of lists per nfil
c
c***************************************************************************
c
      subroutine GETVAL
      include 'tagxref.cmn'
      integer star(5),stop(5)
      character c*1,tag*60,cc*2
c
      if(stri(8:11).ne.'.get')
     * call err(1,"import attribute not recognised ",stri(1:nchr))
      if(d)write(6,'(/a,i3)')' NEW IMPORT list for file',nfil
c
      j=0
      n=0
      call GETTOK
      if(ieof.gt.0)call err(1,'import list misconstructed','')
      if(type.eq.'list')    goto 1
      call err(1,'import attributes expected in table{}',stri(1:nchr))
c
1     k=0
      do 5 i=1,5
      star(i)=0
5     stop(i)=0
10    j=j+1
      if(j.ge.nlis)  goto 50
      c=list(j:j)
      if(c.ne."{")   goto 15
      n=n+1
      k=1
      goto 10
15    if(c.ne.":")   goto 20
      k=2
      goto 10
20    if(c.eq.'}')   goto 40
      if(c.ne.',')   goto 25
      k=1
      goto 10
25    if(c.eq."'")   goto 10
      if(c.eq.'"')   goto 10
      if(c.eq.' ')   goto 10
      if(k.eq.2)     goto 30
      cc=list(j:j+1)
      if(cc.eq.'fi') kk=1
      if(cc.eq.'sa') kk=2
      if(cc.eq.'du') kk=3
      if(cc.eq.'mi') kk=4
      if(cc.eq.'mo') kk=5
      j=j+3
      goto 10
30    if(stop(kk).eq.0) star(kk)=j
      stop(kk)=j+1
      goto 10
c            end of {..}, store arg strings in dicn(), lengths in dicl()
40    continue
cdbg  write(6,'(3i5)') (i,star(i),stop(i),i=1,5)
      do 45 i=1,5
      tag=list(star(i):stop(i)-1)
      call lcase(tag,tag,k)
      dicn(nfil,n,i)=tag(1:k)
45    dicl(nfil,n,i)=stop(i)-star(i)
      if(j.lt.nlis)  goto 1 
c
c ----------------------- put import count per file in nimp()
50    nimp(nfil)=n
      mimp(nfil)=0
      if(d)write(6,'((a,i3,x,2a20,3a7))')('     IMPORT',       
     *  j,(dicn(nfil,j,k),k=1,5),j=1,n)
c
      return
      end
c
c***************************************************************************
c
c    NEXIMP - point to the next import list for this file to process    
c
c***************************************************************************
c
      subroutine NEXIMP
      include 'tagxref.cmn'
c
      next=0
      if(nfil.le.0)       goto 50  
      n=nfil-1
      if(mimp(n).eq.0) goto 20
      m=mimp(n)
      filn=dicn(n,m,1)
      fram=dicn(n,m,2)
      dupl=dicn(n,m,3)
      miss=dicn(n,m,4)
      mode=dicn(n,m,5)
      if(d)write(6,'(a,4i3,x,2a20,x,a4)')' ^^^^^ current file',
     * n,slev(nfil),m,nimp(n),filn,fram,mode
20    if(mimp(nfil).gt.nimp(nfil)) goto 50
      mimp(nfil)=mimp(nfil)+1
      m=mimp(nfil)
      if(m.gt.nimp(nfil)) goto 50
      filn=dicn(nfil,m,1)
      fram=dicn(nfil,m,2)
      dupl=dicn(nfil,m,3)
      miss=dicn(nfil,m,4)
      mode=dicn(nfil,m,5)
      if(d)write(6,'(a,4i3,x,2a20,x,a4)')'  apply import file',
     * nfil,slev(nfil),m,nimp(nfil),filn,fram,mode 
c     
      call NEWFIL
      writ=0
      cach=0
      next=1
c
50    continue
      return
      end
c
c
c***************************************************************************
c
c    GETCAC - get the frame from the cache if its present
c             ncach() contains the store frames
c             icach() contains pointers to the cache line buffer cacbuf()
c
c***************************************************************************
c
      subroutine GETCAC
      include 'tagxref.cmn'
c
      cacy=0
      if(mode.eq.'full'.or.nbuf.eq.0)     goto 80
      if(d)write(6,'(a,a)') ' searching cache for ',fram
      do 65 idef=1,mdef
      if(fram.eq.ncach(idef)) goto 70
65    continue
      if(d)write(6,'(2a20)') fram,' not found in cache '
      goto 80
c                        found in cache; write lines into file7 
70    cacy=1
      do 75 i=icach(idef),icach(idef+1)-2
      len=caclen(i)
      if(d)write(6,'(a,i5,x,a)')' >>>from cache',i,cacbuf(i)(1:len)
75    call WRITLN(cacbuf(i),len)  
c
80    continue
      return
      end
c
c
c***************************************************************************
c
c    TABCAC - tabulate the cache buffer names and positions              
c
c***************************************************************************
c
      subroutine TABCAC
      include 'tagxref.cmn'
c
      cach=0
      writ=1
      if(mode.eq.'full') goto 100
c
      wcan=1
      mdef=mdef+1
      nbuf=nbuf+1
      if(mdef.gt.200) call err(1,'cached defns > 200',fram)
      idef=mdef
      icach(idef+1)=nbuf+1
      ncach(idef)=fram
100   return
      end
c
c***************************************************************************
c
c    NEXFRM - position file at start of the next request import frame  
c
c***************************************************************************
c
      subroutine NEXFRM
      include 'tagxref.cmn'
      character tag*60
c
      n=dicl(nfil-1,mimp(nfil-1),2)
100   call GETTOK
      if(ieof.ne.0) call err(1,'eof searching  ',fram)
      if(type.ne.'save')           goto 100
      if(nchr.le.5)                goto 100
      tag=stri(5:nchr)
      call lcase(tag,tag,l)
      if(r)write(6,'(2a,x,a)')'......searching framecode ',tag,fram
      if(tag(1:l).eq.fram(1:n))    goto 150
      if(tag(2:l).eq.fram(1:n))    goto 150
      goto 100
c                        reset save levels and write flags
150   continue
      call NESLEV
      if(mode.eq.'full') writ=1
      if(mode.ne.'full') cach=1              
c
      return
      end
c
c***************************************************************************
c
c    GETFM1 - copy requested frame to output file7 (xrodic or xrotex)     
c
c***************************************************************************
c
      subroutine GETFM1
      include 'tagxref.cmn'
c 
200   ichr=maxc
      call GETTOK
      if(ieof.ne.0)      goto 220
      writ=1
      if(type.ne.'save') goto 210
      if(d)write(6,'(a,2i3,x,a60)')" save_ in file",nfil,slev(nfil),ibuf
      call NESLEV
      if(send.eq.1)      goto 220
210   if(type.ne.'impt') goto 200
c...........................................
      call JMPORT
c...........................................
      if(ieof.ne.0)      goto 220
      goto 200
c
220   continue
      return
      end
c
c
c***************************************************************************
c
c    GETFM2 - copy requested frame to output file7 (xrodic or xrotex)     
c
c***************************************************************************
c
      subroutine GETFM2
      include 'tagxref.cmn'
c 
200   ichr=maxc
      call GETTOK
      if(ieof.ne.0)      goto 220
      writ=1
      if(type.ne.'save') goto 210
      call NESLEV
      if(d)write(6,'(a,3i3)')"save_ in file",nfil,slev(nfil),send
      if(send.eq.1)      goto 220
210   if(type.ne.'impt') goto 200
c...........................................
      call IMPORT
c...........................................
      if(ieof.ne.0)      goto 220
      goto 200
c
220   continue
      return
      end
c
c***************************************************************************
c
c     JMPORT is an identical version of IMPORT and provides for recursion
c
c***************************************************************************
c
      subroutine JMPORT
      include 'tagxref.cmn'
c
c                        get import args from list [{..},....]
      call GETVAL
c                        reset the import list pointer
100   call NEXIMP
      if(next.eq.1)       goto 200
      if(slev(nfil).gt.0) goto 400
      goto 800
c                        if there is a cache, search & retrieve
200   call GETCAC
      if(cacy.eq.1) goto 100
c                        not in cache find frame in import file
300   call NEXFRM
c                        get the import frame and copy to xrodic/xrotex
400   call GETFM2
      if(ieof.ne.0) goto 1000
c                       tabulate cache pointers for any new lines
      call TABCAC
      goto 100
c                         close import file nfil and decrement
800   if(nfil.le.1) goto 1000
      call CLOSEF
      goto 100
c
1000  writ=1
      if(d)write(6,'(2a,a30)')'\\\\\\  exit IMPORT ',mode,fram
c
      return
      end
c
c***************************************************************************
c     NESLEV - checks saveframe levels as stack for each file.       
c***************************************************************************
      subroutine NESLEV
      include 'tagxref.cmn'
      character*100 dump
c
      send=0
      if(nchr.eq.5) goto 10
c                         open frame - increase level
      slev(nfil)=slev(nfil)+1
      goto 100
c                        close frame - decrease level
10    continue
      slev(nfil)=slev(nfil)-1
      if(slev(nfil).gt.0) goto 100
      send=1
c
100   continue
      if(d)write(6,'(a,3i3)') '```````````NESLEV',nfil,slev(nfil),send
      return
      end
c
c
c***************************************************************************
c     ERR message generator. output to device 6 and stop.
c***************************************************************************
      subroutine ERR(i,mess,strg)
      include 'tagxref.cmn'
      integer i
c
      character*(*) mess,strg
c
      write(*,'(11h error >>> ,a,2x,a)') mess,strg
      write(4,'(11h error >>> ,a,2x,a)') mess,strg
      if(i.eq.2) goto 20
      write(*,'(11x,a,i5,3x,a)') 'at line',line(ndev),filn
      write(4,'(11x,a,i5,3x,a)') 'at line',line(ndev),filn
      if(i.eq.0) return
      goto 50
20    write(*,'(11x,a,i5)') 'in OTEMP at import table level', nfil
      write(4,'(11x,a,i5)') 'in OTEMP at import table level', nfil
50    close(4)
      stop
      end
c
c***************************************************************************
c     lcase converts strings to lower case characters.
c***************************************************************************
c
      subroutine lcase(strin,strot,l)
      include 'tagxref.cmn'
      character strin*60,strot*60,c*1
      integer n,l
c
      n=0
      l=0
10    n=n+1
      if(n.gt.60) goto 100
      c=strin(n:n)
      if(c.ne.' ')  goto 15
      strot(n:n)=' '
      goto 10
15    l=l+1
      if(c.ge.'A'.and.c.le.'Z') goto 20
      strot(n:n)=c
      goto 10
20    strot(n:n)=char(ichar(c)+case)
      goto 10
100   return
      end
c
c***************************************************************************
c     NEWFIL opens a new dictionary file into the stack indexed by nfil.
c***************************************************************************
      subroutine NEWFIL
      include 'tagxref.cmn'
c                                save ip buf of current file
      if(nfil.gt.0) lbuf(nfil)=ibuf
c                                store import history if xritex
10    if(nhis.lt.0) goto 50
      nhis=nhis+1
      if(nfil.eq.0.or.nhis.eq.1) goto 30
      if(nhis.gt.1000)call err(1,'import history table >1000','')
      do 20 i=1,nfil
      do 20 j=1,5
20    hist(nhis,i,j)=hist(nhis-1,i,j)
30    if(nfil+1.gt.25) call err(1,'dict file import stack > 25','')
      hist(nhis,nfil+1,1)=filn
      hist(nhis,nfil+1,2)=fram
      hist(nhis,nfil+1,3)=dupl
      hist(nhis,nfil+1,4)=miss
      hist(nhis,nfil+1,5)=mode
c
50    nfil=nfil+1
      slev(nfil)=0
      if(nfil.gt.mfil) mfil=nfil
      ndev=devn(nfil)
      open(ndev,file=filn,status='old',form='formatted')
      line(ndev)=0
      ichr=maxc
      ieof=0
      nlin=0
      if(h)write(6,'(8(2h> ),a,i4,x,2a20,x,a4)')'open',nfil,filn,fram
      return
      end
c
c***************************************************************************
c     WRITLN writes a line to file7 (xrodic or xrotex)
c***************************************************************************
      subroutine WRITLN(buf,len)
      include 'tagxref.cmn'
c
      character*125 buf
      integer len
c
      write(7,'(a)') buf(1:len)
      if(w)write(6,'(a,i5,x,a)')' writln ++++++',len,buf(1:len)
      return
      end
c
c***************************************************************************
c     RWLINE reads a line from the current file (nfil) on device ndev.
c            write the previous line to file7 if writ=1 & nlin>0 & wcan=0
c            the previous line is cached in cacbuf() if cach = 1.
c***************************************************************************
      subroutine RWLINE
      include 'tagxref.cmn'
c
c     -------------------------------------- write prev buffer to file7 
c
      if(w)write(6,'(a,5i9)')' RWLINE ++++++++++++++++',writ,nlin,wcan   
      if(writ.eq.0.or.nlin.eq.0.or.wcan.gt.0)  goto 20
      write(7,'(a)') ibuf(1:mchr)
      if(w)write(6,'(a,i5,x,a)')' WRITLN ******',mchr,ibuf(1:mchr)
c
c     -------------------------------------- output line to cacbuf()      
10    if(cach.eq.0) goto 20
      nbuf=nbuf+1
      if(nbuf.gt.2000) call err(1,'dbuf() defn cache > limit','2000')
      cacbuf(nbuf)(1:mchr)=ibuf(1:mchr)
      caclen(nbuf)        =mchr
      if(d)write(6,'(a,i5,x,a)') ' to cache <<<<',nbuf,ibuf(1:mchr)
c
c     -------------------------------------- read a new line on ndev      
20    read(ndev,'(a)',end=100) ibuf(1:maxc)
      nlin=nlin+1
      line(ndev)=line(ndev)+1
      do 30 i=maxc,2,-1
30    if(ibuf(i:i).ne.' ') goto 25
25    mchr=i
      ichr=1
      wcan=wcan-1
      if(r)write(6,'(a,i5,x,a)') ' READLN {{{{{{',mchr,ibuf(1:mchr)
      goto 200
c
100   ieof=1
      wcan=0
200   return
      end
c
c***************************************************************************
c     GETTOK extract a TOKEN of characters delimited by blanks, tabs, quotes.
c***************************************************************************
      subroutine GETTOK
      include 'tagxref.cmn'
c
      integer       i,lbra
      character c*1,x*1,z*1,num*13,temp*125,ccc*3
      data num/'0123456789+-.'/
c
c---------------------------------- incr char on current line
10    ichr=ichr+1
      if(ichr.le.mchr)     goto 30
c---------------------------------- get a new line from device ndev
20    call RWLINE
      if(ieof.ne.0) goto 300
c---------------------------------- is this semicolon delimited block
      if(ibuf(1:1).ne.';') goto 30
      if(semi.eq.1) goto 27
      semi=1
      if(text.eq.0) goto 25
      if(b)         goto 10
25    call RWLINE
      if(ieof.ne.0) goto 300
      if(ibuf(1:1).ne.';') goto 25
27    semi=0
      goto 10
c---------------------------------- test for delimiter character
30    c=ibuf(ichr:ichr)
      if(c.eq.' ')       goto 10
      if(c.eq.tab)       goto 10
      if(c.eq.'#')       goto 20
      if(c.eq.'''')      goto 60
      if(c.eq.'"')       goto 60
      if(c.eq.'[')       goto 70
      if(c.eq.'{')       goto 70
c---------------------------------- test for data name
      if(c.ne.'_')       goto 40
      type='name'
      iatt=0
      goto 45
c---------------------------------- test for number or character data
40    type='numb'
      if(index(num,c).eq.0) type='char'
c---------------------------------- get blank-limited char string
45    do 50 i=ichr,mchr
      if(ibuf(i:i).eq.' ')       goto 90
      if(ibuf(i:i).eq.tab)       goto 90
50    continue
      i=mchr+1
      goto 90
c---------------------------------- scan single & triple quoted string
60    type='char'
      ccc=c//c//c
      if(ibuf(ichr:ichr+2).ne.ccc) goto 64
c                                           triple quotes
61    do 62 i=ichr+3, mchr
      if(ibuf(i:i+2).eq.ccc)       goto 63
62    continue
      call RWLINE
      if(ieof.ne.0) goto 300
      ichr=-2
      goto 61
63    i=i+3
      goto 90
c                          i                 single quotes
64    do 65 i=ichr+1,mchr
      if(ibuf(i:i).eq.c)         goto 68
65    continue
      if(dicf.eq.1)
     * call err(0,'quoted string not closed',ibuf(ichr:mchr))
      i=mchr
      ibuf(i:i)=c
68    i=i+1
      goto 90
c---------------------------------- get brace-delimited string
70    if(type.eq.'impt') wcan=10000
      type='list'
      lbra=1
      nlis=1
      if(c.eq.'[') z=']'
      if(c.eq.'{') z='}'
c                                  find balancing open brace [ {
72    j=0
      k=1
      do 76 i=ichr+1,mchr
      x=ibuf(i:i)
      if(x.eq.' ')          goto 76
      if(j.eq.0)  k=i
      j=i
73    if(x.ne.z)            goto 74
      lbra=lbra-1
      if(lbra.eq.0)         goto 78
74    if(x.ne.c)            goto 76
      lbra=lbra+1
76    continue
c                              load list string so far into list()
78    n=nlis+j-k+1
      if(n.gt.10000) goto 250
      list(nlis:n)=ibuf(k:j)
      nlis=n+1
      if(lbra.eq.0) goto 85
      call RWLINE
      if(ieof.ne.0)call err(1,'unexpected eof inputting List','')
      ichr=0
      goto 72
c                              list() has closed; resume parse
85    continue
      wcan=0
      i=j+1
      if(s)write(6,'(2i5,3(2x,a))') nlis,j,c,z,list(1:nlis)
c---------------------------------- count & store string
90    nchr=i-ichr
      ichr=i-1
      stri(1:nchr)=ibuf(i-nchr:ichr)
c----------------------------------- test if char str is save_
      if(stri(1:5).ne.'save_') goto 100
      type='save'
      goto 400
c----------------------------------- test if char str is IMPORT
100   if(stri(1:7).ne.'_import') goto 200
      if(semi.eq.1) goto 400
      type='impt'
      wcan=1
      goto 400
c----------------------------------- test if char str is loop_
200   if(stri(1:5).ne.'loop_') goto 400
      type='loop'
      temp=ibuf
      goto 400
c
250   call err(1,'list >10000ch',' check for balanced braces')
c-----------------------------------eof encountered 
300   continue
      if(nfil.gt.1) goto 400
c                                  only the parent file can exit
c                                  all others must exit by caller
      close(ndev)
      if(nfil.eq.0) goto 400
      if(h) write(6,'(a,i3/)') '---gstr eof---- close',nfil
      nfil=nfil-1
      ichr=maxc
400   continue
      if(s)write(6,'(3(x,a))')'getstr::::',type,stri(1:nchr)
      return
      end
C
C
C
C
C
C
C
C
C
c
c    OCCASIONAL PROGRAMMING NOTES............
c
c    TAGXREF is a recursive program, and as such the process flow appears
c    somewhat convoluted. Here are a few pertinent notes that will help
c    in any future changes. They are as much for the current author as
c    as for any future ones!
c
c    TAGXREF does not currently check any imported definition material for
c    either duplicate save frames or contents of save frames or for that
c    matter, validate for STAR compliance - it assumes that both the dic
c    and text files have been checked with the routine ALIGN.               
c     
c    Because Fortran (or at least the gnu 'gfortran' I am using) will not
c    allow a routine to call itself (i.e. no recursion) this is achieved
c    by subterfuge... the routine IMPORT calls an identical routine JMPORT 
c    which in turn calls IMPORT, and so on. Recursion is necessary because
c    imported lines often contain other import commands and these can be
c    nested to any level. This version of TAGXREF allows for up to 25 levels.
c
c    An encountered import.get list is stored in the character array dicn()
c    and the integer array dicl(), as follows: 
c    dicn(nfil,imp,1) is the filename   entered as "file":'filename'
c    dicn(nfil,imp,2) is the framecode  entered as "save":'filename'
c    dicn(nfil,imp,3) is the importmode entered as "mode":'filename'
c    dicn(nfil,imp,4) is the duplcode   entered as "dupl":'filename'
c    dicn(nfil,imp,5) is the missing op entered as "miss":'filename'
c    dicl(nfil,imp,1-5) are the lengths of the input strings
c   
c    'imp' are the separate import requests in an import.get list. See
c     further notes on this below with mimp() and nimp().
c
c    'nfil' is the current file being processed for imports. For example,
c    the input dictionary 'idict' is always nfil=1. If an import.get []
c    is encountered in idict the requested import file is immediately
c    opened as nfil=2 and the requested frame is written to odict. If
c    within this frame another import request is encountered (note that 
c    any import.get's in the file outside the requested frame are ignored) 
c    another file is opened and importation proceeds at the next level 
c    with nfil=3. In turn, additional files are opened if imports are 
c    encountered within these import frames, accordingly nfil values of 
c    4,5...25 are assigned. When importation of a frame is completed
c    the current file being processed is closed and nfil is decremented.
c    the process then applies to that nfil value until its frame is 
c    imported.... and so on until nfil=1 and the last line of 'xridic'
c    has been copied to 'xrodic'. That is, all lines in 'xridic' are copied
c    to 'xrodic' but with the import.get's replaced with the requested
c    imported lines.
c
c    If 'xritex' is also a dictionary file, the identical process to the 
c    above is applied and the file 'xrotex' is generated. The defined data
c    names in 'xrodic' are searched for in 'xrotex'.
c
c    It is worth noting the function of several other parameters as they
c    play a crucial role in the recursive import process.
c
c      slev(nfil) is the current save frame level of the import frame.
c                 Until the requested frame is found in nfil its value 
c                 remains at 0. When the frame is found its value is 1
c                 and with every opening save frame it is incremented 
c                 (in NESLEV) and with every closing save frame it is
c                 decremented. The import file nfil is only closed
c                 when slev(nfil) returns to 0. Because of the recursion
c                 many slev values can be non-zero at the same time. 
c                 slev(1) is always 0 because it is not being imported
c                 and when the current nfil being processed is 1, then
c                 all slev values will be 0.
c
c     mimp(nfil) is the current import request 'imp' applied to nfil.
c
c     nimp(nfil) is the total number imports requested for this nfil.
c                mimp(nfil) is incremented each time slev(nimp+1) returns
c                to 0 (and file nfil+1 is closed) and that when mimp(nfil)
c                exceeds nimp(nfil) the processing of nfil resumes until
c                its slev(nfil) returns to 0, and so on....
c
c     A crucial aspect of the TAGXREF algorithm to appreciate is that the 
c     import request information in dicn(nfil), dicl(), mimp(), mimp() 
c     refers to the nfil` value of the parent file (i.e. the file in which 
c     the .get is encountered) ... but the file opened to retrieve the 
c     requested frame has the nfil value =nfil`+1. NB the slev() values
c     refer to the incremented value nfil`+1!
c
c     Happy coding!   .......SRH
